# dbdbd -- David Black's DataBase Definer, version 0.2.2
#          Copyright (c) 2002, 2003 David Alan Black
#                              dblack@wobblini.net
#          See README for license information.

# ./dbdbd.rb
# $Date: 2003-12-26 23:04:43 -0500 (Fri, 26 Dec 2003) $
# $Rev: 103 $
# $Author: dblack $

require "scanf"

# Collection of utility modules and classes.
module DBDBD

  # Regular expressions for data parsing.
  module Re
    BLANK     = /^\s*$/
    COMMENT   = /^\s*#/
    DATA      = /^\s*[^\s#]/
    ENDDATA   = /^END$/
    FIELDS    = /^\s*#(.*)/
  end

  # Make an object respond to "-".
  module MinusStub
    def -(val)
      self
    end
  end

  # scanf-style format string
  class FormatString < String
    # How many fields will this string generate?
    def field_count
      count("%")
    end

    # Break the format string into specifier substrings.
    def specs
      scan(/%[^%]+/)
    end
  end

  # Container for header (comment block at top of file)
  class Header < Array
    def process
      self
    end
  end
  
  # Line containing field-names.
  class Fieldline < String
    def initialize
      super("")
    end
    def process
      split(/\s+/)[1..-1] || []
    end
  end

  # Buffer for data (record rows and comments)
  class RawData < Array
    # Pick comment/data sequences out of raw data from file.
    def parse
      cmt = ""
      each do |s|
	case s
	when Re::ENDDATA  then yield cmt
	when Re::DATA     then yield cmt,s; cmt = ""
	when Re::COMMENT  then cmt << s
	end
      end
      yield cmt
    end
  end

  # Everything after /^END$/ line
  class EndMatter < Array
    def process
      self
    end
  end
  
  # Container for malformed lines (output to endmatter)
  class Malformed < Array
    def to_str
      if empty?
	""
      else
      <<-EOS
      --
Malformed entries (from this or earlier runs; must be deleted manually)

#{join("--\n")}
EOS
      end
    end
  end

  # Staging area prior to writing database.
  class Buffer < Array
  end
  
  # This is the basic Record class.  It gets subclassed by a new Record class
  # for each new database.
  class Record < Array
    attr_reader :db
    attr_reader :comment
    
    # Create a new record, passing in an array of values.
    def initialize(data)
      @db = self.class.db
      replace data
      super
    end
    
    alias :oldset :[]=
      
    # Fetch a field by key.
    def [](key)
      at(db.fields.index(key) || key)
    end
    
    # Stub for no-op strip method.
    module Stripper
      def strip
	self
      end
    end

    # Set a field.
    def []=(field,val)
      val.extend(Stripper) unless val.respond_to?(:strip)
      oldset(db.fields.index(field) || field,val.strip)
    end
    
    # Set the comment for a record
    def comment=(str_or_ary)
      @comment = Comment[str_or_ary]
    end
  end

  # File class with some DBDB-specific intelligence.
  class File < File
    # Rewind by length of string.
    def ungets(str)
      seek(pos - str.size,0)
    end

    # One-line lookahead.
    def next_line
      line = gets
      ungets(line) if line
      line
    end

    # Read header (comment block at top of file)
    def get_header
      h = Header.new
      h << gets until Re::DATA.match(h[-1]) || eof?
      unless eof?
	[ 
	  [Re::DATA,   h[-1]], 
	  [Re::FIELDS, h[-2]] ].each do |re,str|
	  if re.match(str)
	    h.pop
	    ungets(str)
	  end
	end
      end
      h
    end

    # Read line of field-names.
    def get_fieldline
      f = Fieldline.new
      if Re::FIELDS.match(next_line || "")
	f.replace(gets)
      end
      f
    end

    # Read data (record rows and comments).
    def get_data
      res = []
      res << (line=gets) until ((e = Re::ENDDATA.match(line)) || eof?)
      ungets(line) if e
      RawData.new.replace(res.find_all {|r| /\S/.match(r)})
    end

    # Read what's after the data.
    def get_endmatter
      EndMatter.new.replace(readlines)
    end
  end

  # Comment-string manipulation.
  class Comment
    # Format a (possibly already formatted) string or array into
    # comment form.
    def self.[](str_or_ary)
      str_or_ary.to_a.map {|b| b.sub(/(^# )?(.*)\n?/, "# \\2\n")}
    end
  end
end

class DBDB < Hash
  attr_accessor :header, :fields
  attr_reader :endmatter
  attr_writer :file

  alias :records :values
  alias :tie :file=

  # Pass in a scanf format string and optional array of field names.
  def initialize(fmt, *fields)
    @fmt           = DBDBD::FormatString.new(fmt)
    @fields        = fields
    self.sort_key  = 1
    self.uniq_key  = 1
    @malformed     = DBDBD::Malformed.new
    @rc            = Class.new(DBDBD::Record)
    class << @rc;  attr_accessor :db; end

    @rc.db = self
  end

  # Read the file into the database.
  def read
    dbf            = DBDBD::File.new(@file, "r")
    @header        = dbf.get_header.process
    @fields        = dbf.get_fieldline.process
    bulk_insert      dbf.get_data
    @endmatter     = dbf.get_endmatter.process
  end

  # Write the database to file.
  def sync
    File.open(@file, "w") {|fh| fh.print(prep_db)}
  end

  # Write the database to STDOUT.
  def dump
    STDOUT.print prep_db
  end
  
  # Write the database to file and clear it.
  def sync!
    begin; sync; rescue; raise; end
    clear
  end

  # Write the database to STDOUT and clear it.
  def dump!
    begin; dump; rescue; raise; end
    clear
  end

  # Add a new Record to the database.
  def insert(*args)
    i = fields.index(@uniq_key) || @uniq_key
    self[args[i]] = @rc.new(args)
  end

  # Set the database header (comment block at the top)
  def header=(str_or_ary)
    @header = DBDBD::Comment[str_or_ary]
  end

  def endmatter=(str_or_ary)
    DBDBD::EndMatter.new.replace(str_or_ary.to_a)
  end

  # Set sort key (1-originating)
  def sort_key=(val)
    @sort_key = minus_key(val)
  end

  # Set unique key (1-originating)
  def uniq_key=(val)
    @uniq_key = minus_key(val)
  end
    
  # uniq_key (field name or 1-originating index)
  def uniq_key
    @uniq_key + 1
  end

  # sort_key (field name or 1-originating index)
  def sort_key
    @sort_key + 1
  end

  private

  # Masquerade arrays as 1-originating,
  # also allowing for string values.
  def minus_key(val)
    val.extend(DBDBD::MinusStub) unless val.respond_to?(:-)
    val - 1
  end
    
  # Add many records, including comments, from raw data.
  def bulk_insert(data)
    data.parse do |cmt,line|
      if cmt &&! line
	@stray_comment = cmt
      else
	items = line.strip.scanf(@fmt + "%s").map do |it|
	  if it.respond_to?(:strip) then it.strip else it end
	end
	if items.size != @fmt.field_count
	  puts "Warning: malformed entry: #{line}"
	  @malformed << cmt.to_s + line
	else
	  insert(*items) .comment = cmt || ""
	end
      end
    end
  end

  # Sort records prior to output.  sort_key is user-settable.
  def sort_proc
    proc {|a,b| a[1][@sort_key] <=> b[1][@sort_key]}
  end

  # Format the whole database for print or file output.
  def prep_db
    sfmt = make_output_format
    [
      @header,
      sprintf(field_line, @fields),
      sort(&sort_proc).map do |k,rec| 
	[rec.comment, make_line(sfmt,rec)]
      end,
      @stray_comment,
      @endmatter,
      @malformed
    ] .to_s
  end

  # Field names supplemented by numerical pseudo-fieldnames.
  def dummy_fields
    (0...@fmt.field_count).map {|i| fields[i] || i.to_s}
  end

  # Manufacture a "%s" sprintf specifier with length.
  def s_spec(len)
    "%-#{len}s"
  end

  # Create an sprintf format string, using fixed widths or
  # finding largest element in each column.
  def make_output_format
    of = (0...dummy_fields.size).map do |c|
      s_spec(
	     @fmt.specs[c].scan(/\d+/)[0] ||
	     biggest_in_column(c) + 2
	     )
    end .to_s + "\n"
  end
  
  # Find the biggest item in a given column, including the fieldname.
  def biggest_in_column(c)
    [dummy_fields[c].size, *records.map {|r| r[c].to_s.size}] .max
  end

  # Generate the field line for output
  def field_line
    make_line(make_output_format,fields,"# ")
  end

  # Generic method to make a line based on a format string.
  def make_line(sfmt, elements, pre="  ")
    if elements.empty?
      ""
    else
      sprintf(pre + sfmt, *elements).sub(/\s*$/,'') + "\n"
    end
  end

end
